import { Selector } from "./types";
/**
 * Turns `selector` back into a string.
 *
 * @param selector Selector to stringify.
 */
export declare function stringify(selector: Selector[][]): string;
//# sourceMappingURL=stringify.d.ts.map